#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This is a small example which creates a twitch stream to connect with
and changes the color of the video according to the colors provided in
the chat.
"""
from __future__ import print_function
from twitchstream.outputvideo import TwitchBufferedOutputStream
from twitchstream.chat import TwitchChatStream
import argparse
import time
import numpy as np

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    required = parser.add_argument_group('required arguments')
    required.add_argument('-u', '--username',
                          help='twitch username',
                          required=True)
    required.add_argument('-o', '--oauth',
                          help='twitch oauth '
                               '(visit https://twitchapps.com/tmi/ '
                               'to create one for your account)',
                          required=True)
    required.add_argument('-s', '--streamkey',
                          help='twitch streamkey',
                          required=True)
    args = parser.parse_args()

    # load two streams:
    # * one stream to send the video
    # * one stream to interact with the chat
    with TwitchBufferedOutputStream(
            twitch_stream_key=args.streamkey,
            width=640,
            height=480,
            fps=30.,
            enable_audio=True,
            verbose=False) as videostream, \
        TwitchChatStream(
            username=args.username,
            oauth=args.oauth,
            verbose=False) as chatstream:

        # Send a chat message to let everybody know you've arrived
        chatstream.send_chat_message("Taking requests!")

        frame = np.zeros((480, 640, 3))
        frequency = 100
        last_phase = 0

        # The main loop to create videos
        while True:

            # Every loop, call to receive messages.
            # This is important, when it is not called,
            # Twitch will automatically log you out.
            # This call is non-blocking.
            received = chatstream.twitch_receive_messages()

            # process all the messages
            if received:
                for chat_message in received:
                    print("Got a message '%s' from %s" % (
                        chat_message['message'],
                        chat_message['username']
                    ))
                    if chat_message['message'] == "red":
                        frame[:, :, :] = np.array(
                            [1, 0, 0])[None, None, :]
                    elif chat_message['message'] == "green":
                        frame[:, :, :] = np.array(
                            [0, 1, 0])[None, None, :]
                    elif chat_message['message'] == "blue":
                        frame[:, :, :] = np.array(
                            [0, 0, 1])[None, None, :]
                    elif chat_message['message'].isdigit():
                        frequency = int(chat_message['message'])

            # If there are not enough video frames left,
            # add some more.
            if videostream.get_video_frame_buffer_state() < 30:
                videostream.send_video_frame(frame)

            # If there are not enough audio fragments left,
            # add some more, but take care to stay in sync with
            # the video! Audio and video buffer separately,
            # so they will go out of sync if the number of video
            # frames does not match the number of audio samples!
            elif videostream.get_audio_buffer_state() < 30:
                x = np.linspace(last_phase,
                                last_phase +
                                frequency*2*np.pi/videostream.fps,
                                int(44100 / videostream.fps) + 1)
                last_phase = x[-1]
                audio = np.sin(x[:-1])
                videostream.send_audio(audio, audio)

            # If nothing is happening, it is okay to sleep for a while
            # and take some pressure of the CPU. But not too long, if
            # the buffers run dry, audio and video will go out of sync.
            else:
                time.sleep(.001)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
This file contains the python code used to interface with the Twitch
chat. Twitch chat is IRC-based, so it is basically an IRC-bot, but with
special features for Twitch, such as congestion control built in.
"""
from __future__ import print_function
import time
import socket
import re
import fcntl
import os
import errno


class TwitchChatStream(object):
    """
    The TwitchChatStream is used for interfacing with the Twitch chat of
    a channel. To use this, an oauth-account (of the user chatting)
    should be created. At the moment of writing, this can be done here:
    https://twitchapps.com/tmi/

    :param username: Twitch username
    :type username: string
    :param oauth: oauth for logging in (see https://twitchapps.com/tmi/)
    :type oauth: string
    :param verbose: show all stream messages on stdout (for debugging)
    :type verbose: boolean
    """

    def __init__(self, username, oauth, verbose=False):
        """Create a new stream object, and try to connect."""
        self.username = username
        self.oauth = oauth
        self.verbose = verbose
        self.current_channel = ""
        self.last_sent_time = time.time()
        self.buffer = []
        self.s = None

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, type, value, traceback):
        self.s.close()

    @staticmethod
    def _logged_in_successful(data):
        """
        Test the login status from the returned communication of the
        server.

        :param data: bytes received from server during login
        :type data: list of bytes

        :return boolean, True when you are logged in.
        """
        if re.match(r'^:(testserver\.local|tmi\.twitch\.tv)'
                    r' NOTICE \* :'
                    r'(Login unsuccessful|Error logging in)*$',
                    data.strip()):
            return False
        else:
            return True

    @staticmethod
    def _check_has_ping(data):
        """
        Check if the data from the server contains a request to ping.

        :param data: the byte string from the server
        :type data: list of bytes
        :return: True when there is a request to ping, False otherwise
        """
        return re.match(
            r'^PING :tmi\.twitch\.tv$', data)

    @staticmethod
    def _check_has_channel(data):
        """
        Check if the data from the server contains a channel switch.

        :param data: the byte string from the server
        :type data: list of bytes
        :return: Name of channel when new channel, False otherwise
        """
        return re.findall(
            r'^:[a-zA-Z0-9_]+\![a-zA-Z0-9_]+@[a-zA-Z0-9_]+'
            r'\.tmi\.twitch\.tv '
            r'JOIN #([a-zA-Z0-9_]+)$', data)

    @staticmethod
    def _check_has_message(data):
        """
        Check if the data from the server contains a message a user
        typed in the chat.

        :param data: the byte string from the server
        :type data: list of bytes
        :return: returns iterator over these messages
        """
        return re.match(r'^:[a-zA-Z0-9_]+\![a-zA-Z0-9_]+@[a-zA-Z0-9_]+'
                        r'\.tmi\.twitch\.tv '
                        r'PRIVMSG #[a-zA-Z0-9_]+ :.+$', data)

    def connect(self):
        """
        Connect to Twitch
        """

        # Do not use non-blocking stream, they are not reliably
        # non-blocking
        # s.setblocking(False)
        # s.settimeout(1.0)

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        connect_host = "irc.twitch.tv"
        connect_port = 6667
        try:
            s.connect((connect_host, connect_port))
        except (Exception, IOError):
            print("Unable to create a socket to %s:%s" % (
                connect_host,
                connect_port))
            raise  # unexpected, because it is a blocking socket

        # Connected to twitch
        # Sending our details to twitch...
        s.send(('PASS %s\r\n' % self.oauth).encode('utf-8'))
        s.send(('NICK %s\r\n' % self.username).encode('utf-8'))
        if self.verbose:
            print('PASS %s\r\n' % self.oauth)
            print('NICK %s\r\n' % self.username)

        received = s.recv(1024).decode()
        if self.verbose:
            print(received)
        if not TwitchChatStream._logged_in_successful(received):
            # ... and they didn't accept our details
            raise IOError("Twitch did not accept the username-oauth "
                          "combination")
        else:
            # ... and they accepted our details
            # Connected to twitch.tv!
            # now make this socket non-blocking on the OS-level
            fcntl.fcntl(s, fcntl.F_SETFL, os.O_NONBLOCK)
            if self.s is not None:
                self.s.close()  # close the previous socket
            self.s = s          # store the new socket
            self.join_channel(self.username)

            # Wait until we have switched channels
            while self.current_channel != self.username:
                self.twitch_receive_messages()

    def _push_from_buffer(self):
        """
        Push a message on the stack to the IRC stream.
        This is necessary to avoid Twitch overflow control.
        """
        if len(self.buffer) > 0:
            if time.time() - self.last_sent_time > 5:
                try:
                    message = self.buffer.pop(0)
                    self.s.send(message.encode('utf-8'))
                    if self.verbose:
                        print(message)
                finally:
                    self.last_sent_time = time.time()

    def _send(self, message):
        """
        Send a message to the IRC stream

        :param message: the message to be sent.
        :type message: string
        """
        if len(message) > 0:
            self.buffer.append(message + "\n")

    def _send_pong(self):
        """
        Send a pong message, usually in reply to a received ping message
        """
        self._send("PONG")

    def join_channel(self, channel):
        """
        Join a different chat channel on Twitch.
        Note, this function returns immediately, but the switch might
        take a moment

        :param channel: name of the channel (without #)
        """
        self.s.send(('JOIN #%s\r\n' % channel).encode('utf-8'))
        if self.verbose:
            print('JOIN #%s\r\n' % channel)

    def send_chat_message(self, message):
        """
        Send a chat message to the server.

        :param message: String to send (don't use \\n)
        """
        self._send("PRIVMSG #{0} :{1}".format(self.username, message))

    def _parse_message(self, data):
        """
        Parse the bytes received from the socket.

        :param data: the bytes received from the socket
        :return:
        """
        if TwitchChatStream._check_has_ping(data):
            self._send_pong()
        if TwitchChatStream._check_has_channel(data):
            self.current_channel = \
                TwitchChatStream._check_has_channel(data)[0]

        if TwitchChatStream._check_has_message(data):
            return {
                'channel': re.findall(r'^:.+![a-zA-Z0-9_]+'
                                      r'@[a-zA-Z0-9_]+'
                                      r'.+ '
                                      r'PRIVMSG (.*?) :',
                                      data)[0],
                'username': re.findall(r'^:([a-zA-Z0-9_]+)!', data)[0],
                'message': re.findall(r'PRIVMSG #[a-zA-Z0-9_]+ :(.+)',
                                      data)[0]
            }
        else:
            return None

    def twitch_receive_messages(self):
        """
        Call this function to process everything received by the socket
        This needs to be called frequently enough (~10s) Twitch logs off
        users not replying to ping commands.

        :return: list of chat messages received. Each message is a dict
            with the keys ['channel', 'username', 'message']
        """
        self._push_from_buffer()
        result = []
        while True:
            # process the complete buffer, until no data is left no more
            try:
                msg = self.s.recv(4096).decode()     # NON-BLOCKING RECEIVE!
            except socket.error as e:
                err = e.args[0]
                if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
                    # There is no more data available to read
                    return result
                else:
                    # a "real" error occurred
                    # import traceback
                    # import sys
                    # print(traceback.format_exc())
                    # print("Trying to recover...")
                    self.connect()
                    return result
            else:
                if self.verbose:
                    print(msg)
                rec = [self._parse_message(line)
                       for line in filter(None, msg.split('\r\n'))]
                rec = [r for r in rec if r]     # remove Nones
                result.extend(rec)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Tests for chat.py
"""


def test_logged_in_successful():
    """
    Test TwitchChatStream._logged_in_successful
    """
    from twitchstream.chat import TwitchChatStream
    res = TwitchChatStream._logged_in_successful(
        ":tmi.twitch.tv NOTICE * :Error logging in")
    assert res is False
    res = TwitchChatStream._logged_in_successful(
        ":tmi.twitch.tv NOTICE * :Error logging in\n")
    assert res is False
    res = TwitchChatStream._logged_in_successful(
        ":tmi.twitch.tv NOTICE * :Error logging in\r\n")
    assert res is False
    res = TwitchChatStream._logged_in_successful(
        ":tmi.twitch.tv 001 sdsd :Welcome, GLHF!")
    assert res is True
    res = TwitchChatStream._logged_in_successful(
        ":tmi.twitch.tv 001 sdsd :Your host is tmi.twitch.tv")
    assert res is True
    res = TwitchChatStream._logged_in_successful(
        ":tmi.twitch.tv 001 sdsd :This server is rather new")
    assert res is True
    res = TwitchChatStream._logged_in_successful(
        ":tmi.twitch.tv 001 sdsd :-!")
    assert res is True
    res = TwitchChatStream._logged_in_successful(
        ":tmi.twitch.tv 001 sdsd :You are in a maze of twisty passages,"
        " all alike.")
    assert res is True
    res = TwitchChatStream._logged_in_successful(
        ":tmi.twitch.tv 001 sdsd :>")
    assert res is True



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Configuration of the testing
"""
import pytest


def pytest_addoption(parser):
    """
    Add options for testing
    :param parser:
    :return:
    """
    parser.addoption("--runslow",
                     action="store_true",
                     help="run slow tests")


def pytest_runtest_setup(item):
    """
    Skip the slow tests when running with --runslow
    :param item:
    :return:
    """
    if 'slow' in item.keywords and \
            not item.config.getoption("--runslow"):
        pytest.skip("need --runslow option to run")



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This is a small example which creates a twitch stream to connect with
and changes the color of the video according to the colors provided in
the chat.
"""
from __future__ import print_function
from twitchstream.outputvideo import TwitchBufferedOutputStream
import argparse
import numpy as np

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    required = parser.add_argument_group('required arguments')
    required.add_argument('-s', '--streamkey',
                          help='twitch streamkey',
                          required=True)
    args = parser.parse_args()

    with TwitchBufferedOutputStream(
            twitch_stream_key=args.streamkey,
            width=640,
            height=480,
            fps=30.,
            verbose=True,
            enable_audio=True) as videostream:

        frame = np.zeros((480, 640, 3))

        while True:
            if videostream.get_video_frame_buffer_state() < 30:
                frame = np.random.rand(480, 640, 3)
                videostream.send_video_frame(frame)

            if videostream.get_audio_buffer_state() < 30:
                left_audio = np.random.randn(1470)
                right_audio = np.random.randn(1470)
                videostream.send_audio(left_audio, right_audio)




#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This is a small example which connects to a user's chat channel to
send and receive the messages posted there
"""
from __future__ import print_function
from twitchstream.outputvideo import TwitchOutputStreamRepeater
from twitchstream.chat import TwitchChatStream
import argparse
import time
import numpy as np

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    required = parser.add_argument_group('required arguments')
    required.add_argument('-u', '--username',
                          help='twitch username',
                          required=True)
    required.add_argument('-o', '--oauth',
                          help='twitch oauth '
                               '(visit https://twitchapps.com/tmi/ '
                               'to create one for your account)',
                          required=True)
    args = parser.parse_args()

    # Launch a verbose (!) twitch stream
    with TwitchChatStream(username=args.username,
                          oauth=args.oauth,
                          verbose=True) as chatstream:

        # Send a message to this twitch stream
        chatstream.send_chat_message("I'm reading this!")

        # Continuously check if messages are received (every ~10s)
        # This is necessary, if not, the chat stream will close itself
        # after a couple of minutes (due to ping messages from twitch)
        while True:
            received = chatstream.twitch_receive_messages()
            if received:
                print("received:", received)
            time.sleep(1)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
This file contains the classes used to send videostreams to Twitch
"""
from __future__ import print_function, division
import numpy as np
import subprocess
import signal
import threading
import sys
try:
    import Queue as queue
except ImportError:
    import queue
import time
import os

import requests

AUDIORATE = 44100


class TwitchOutputStream(object):
    """
    Initialize a TwitchOutputStream object and starts the pipe.
    The stream is only started on the first frame.

    :param twitch_stream_key:
    :type twitch_stream_key:
    :param width: the width of the videostream (in pixels)
    :type width: int
    :param height: the height of the videostream (in pixels)
    :type height: int
    :param fps: the number of frames per second of the videostream
    :type fps: float
    :param enable_audio: whether there will be sound or not
    :type enable_audio: boolean
    :param ffmpeg_binary: the binary to use to create a videostream
        This is usually ffmpeg, but avconv on some (older) platforms
    :type ffmpeg_binary: String
    :param verbose: show ffmpeg output in stdout
    :type verbose: boolean
    """
    def __init__(self,
                 twitch_stream_key,
                 width=640,
                 height=480,
                 fps=30.,
                 ffmpeg_binary="ffmpeg",
                 enable_audio=False,
                 verbose=False):
        self.twitch_stream_key = twitch_stream_key
        self.width = width
        self.height = height
        self.fps = fps
        self.ffmpeg_process = None
        self.audio_pipe = None
        self.ffmpeg_binary = ffmpeg_binary
        self.verbose = verbose
        self.audio_enabled = enable_audio
        try:
            self.reset()
        except OSError:
            print("There seems to be no %s available" % ffmpeg_binary)
            if ffmpeg_binary == "ffmpeg":
                print("ffmpeg can be installed using the following"
                      "commands")
                print("> sudo add-apt-repository "
                      "ppa:mc3man/trusty-media")
                print("> sudo apt-get update && "
                      "sudo apt-get install ffmpeg")
            sys.exit(1)

    def reset(self):
        """
        Reset the videostream by restarting ffmpeg
        """

        if self.ffmpeg_process is not None:
            # Close the previous stream
            try:
                self.ffmpeg_process.send_signal(signal.SIGINT)
            except OSError:
                pass

        command = []
        command.extend([
            self.ffmpeg_binary,
            '-loglevel', 'verbose',
            '-y',       # overwrite previous file/stream
            # '-re',    # native frame-rate
            '-analyzeduration', '1',
            '-f', 'rawvideo',
            '-r', '%d' % self.fps,  # set a fixed frame rate
            '-vcodec', 'rawvideo',
            # size of one frame
            '-s', '%dx%d' % (self.width, self.height),
            '-pix_fmt', 'rgb24',  # The input are raw bytes
            '-thread_queue_size', '1024',
            '-i', '-',  # The input comes from a pipe

            # Twitch needs to receive sound in their streams!
            # '-an',            # Tells FFMPEG not to expect any audio
        ])
        if self.audio_enabled:
            command.extend([
                '-ar', '%d' % AUDIORATE,
                '-ac', '2',
                '-f', 's16le',
                '-thread_queue_size', '1024',
                '-i', '/tmp/audiopipe'
            ])
        else:
            command.extend([
                '-f', 'lavfi',
                '-i', 'anullsrc=channel_layout=stereo:sample_rate=44100'
            ])
        command.extend([
            # VIDEO CODEC PARAMETERS
            '-vcodec', 'libx264',
            '-r', '%d' % self.fps,
            '-b:v', '3000k',
            '-s', '%dx%d' % (self.width, self.height),
            '-preset', 'faster', '-tune', 'zerolatency',
            '-crf', '23',
            '-pix_fmt', 'yuv420p',
            # '-force_key_frames', r'expr:gte(t,n_forced*2)',
            '-minrate', '3000k', '-maxrate', '3000k',
            '-bufsize', '12000k',
            '-g', '60',     # key frame distance
            '-keyint_min', '1',
            # '-filter:v "setpts=0.25*PTS"'
            # '-vsync','passthrough',

            # AUDIO CODEC PARAMETERS
            '-acodec', 'libmp3lame', '-ar', '44100', '-b:a', '160k',
            # '-bufsize', '8192k',
            '-ac', '1',
            # '-acodec', 'aac', '-strict', 'experimental',
            # '-ab', '128k', '-ar', '44100', '-ac', '1',
            # '-async','44100',
            # '-filter_complex', 'asplit', #for audio sync?

            # STORE THE VIDEO PARAMETERS
            # '-vcodec', 'libx264', '-s', '%dx%d'%(width, height),
            # '-preset', 'libx264-fast',
            # 'my_output_videofile2.avi'

            # MAP THE STREAMS
            # use only video from first input and only audio from second
            '-map', '0:v', '-map', '1:a',

            # NUMBER OF THREADS
            '-threads', '2',

            # STREAM TO TWITCH
            '-f', 'flv', self.get_closest_ingest(),
        ])

        devnullpipe = subprocess.DEVNULL
        if self.verbose:
            devnullpipe = None
        self.ffmpeg_process = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stderr=devnullpipe,
            stdout=devnullpipe)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        # sigint so avconv can clean up the stream nicely
        self.ffmpeg_process.send_signal(signal.SIGINT)
        # waiting doesn't work because of reasons I don't know
        # self.pipe.wait()

    def send_video_frame(self, frame):
        """Send frame of shape (height, width, 3)
        with values between 0 and 1.
        Raises an OSError when the stream is closed.

        :param frame: array containing the frame.
        :type frame: numpy array with shape (height, width, 3)
            containing values between 0.0 and 1.0
        """

        assert frame.shape == (self.height, self.width, 3)

        frame = np.clip(255*frame, 0, 255).astype('uint8')
        try:
            self.ffmpeg_process.stdin.write(frame.tostring())
        except OSError:
            # The pipe has been closed. Reraise and handle it further
            # downstream
            raise

    def send_audio(self, left_channel, right_channel):
        """Add the audio samples to the stream. The left and the right
        channel should have the same shape.
        Raises an OSError when the stream is closed.

        :param left_channel: array containing the audio signal.
        :type left_channel: numpy array with shape (k, )
            containing values between -1.0 and 1.0. k can be any integer
        :param right_channel: array containing the audio signal.
        :type right_channel: numpy array with shape (k, )
            containing values between -1.0 and 1.0. k can be any integer
        """
        if self.audio_pipe is None:
            if not os.path.exists('/tmp/audiopipe'):
                os.mkfifo('/tmp/audiopipe')
            self.audio_pipe = os.open('/tmp/audiopipe', os.O_WRONLY)

        assert len(left_channel.shape) == 1
        assert left_channel.shape == right_channel.shape

        frame = np.column_stack((left_channel, right_channel)).flatten()

        frame = np.clip(32767*frame, -32767, 32767).astype('int16')
        try:
            os.write(self.audio_pipe, frame.tostring())
        except OSError:
            # The pipe has been closed. Reraise and handle it further
            # downstream
            raise

    def get_closest_ingest(self):
        closest_server = requests.get(url='https://ingest.twitch.tv/api/v2/ingests').json()['ingests'][0]
        url_template = closest_server['url_template']
        print("Streaming to closest server: %s at %s" % (closest_server['name'],
                                                         url_template.replace('/app/{stream_key}', '')))
        return url_template.format(
            stream_key=self.twitch_stream_key)


class TwitchOutputStreamRepeater(TwitchOutputStream):
    """
    This stream makes sure a steady framerate is kept by repeating the
    last frame when needed.

    Note: this will not generate a stable, stutter-less stream!
     It does not keep a buffer and you cannot synchronize using this
     stream. Use TwitchBufferedOutputStream for this.
    """
    def __init__(self, *args, **kwargs):
        super(TwitchOutputStreamRepeater, self).__init__(*args, **kwargs)

        self.lastframe = np.ones((self.height, self.width, 3))
        self._send_last_video_frame()   # Start sending the stream

        if self.audio_enabled:
            # some audible sine waves
            xl = np.linspace(0.0, 10*np.pi, int(AUDIORATE/self.fps) + 1)[:-1]
            xr = np.linspace(0.0, 100*np.pi, int(AUDIORATE/self.fps) + 1)[:-1]
            self.lastaudioframe_left = np.sin(xl)
            self.lastaudioframe_right = np.sin(xr)
            self._send_last_audio()   # Start sending the stream

    def _send_last_video_frame(self):
        try:
            super(TwitchOutputStreamRepeater,
                  self).send_video_frame(self.lastframe)
        except OSError:
            # stream has been closed.
            # This function is still called once when that happens.
            pass
        else:
            # send the next frame at the appropriate time
            threading.Timer(1./self.fps,
                            self._send_last_video_frame).start()

    def _send_last_audio(self):
        try:
            super(TwitchOutputStreamRepeater,
                  self).send_audio(self.lastaudioframe_left,
                                   self.lastaudioframe_right)
        except OSError:
            # stream has been closed.
            # This function is still called once when that happens.
            pass
        else:
            # send the next frame at the appropriate time
            threading.Timer(1./self.fps,
                            self._send_last_audio).start()

    def send_video_frame(self, frame):
        """Send frame of shape (height, width, 3)
        with values between 0 and 1.

        :param frame: array containing the frame.
        :type frame: numpy array with shape (height, width, 3)
            containing values between 0.0 and 1.0
        """
        self.lastframe = frame

    def send_audio(self, left_channel, right_channel):
        """Add the audio samples to the stream. The left and the right
        channel should have the same shape.

        :param left_channel: array containing the audio signal.
        :type left_channel: numpy array with shape (k, )
            containing values between -1.0 and 1.0. k can be any integer
        :param right_channel: array containing the audio signal.
        :type right_channel: numpy array with shape (k, )
            containing values between -1.0 and 1.0. k can be any integer
        """
        self.lastaudioframe_left = left_channel
        self.lastaudioframe_right = right_channel


class TwitchBufferedOutputStream(TwitchOutputStream):
    """
    This stream makes sure a steady framerate is kept by buffering
    frames. Make sure not to have too many frames in buffer, since it
    will increase the memory load considerably!

    Adding frames is thread safe.
    """
    def __init__(self, *args, **kwargs):
        super(TwitchBufferedOutputStream, self).__init__(*args, **kwargs)
        self.last_frame = np.ones((self.height, self.width, 3))
        self.last_frame_time = None
        self.next_video_send_time = None
        self.frame_counter = 0
        self.q_video = queue.PriorityQueue()

        # don't call the functions directly, as they block on the first
        # call
        self.t = threading.Timer(0.0, self._send_video_frame)
        self.t.daemon = True
        self.t.start()

        if self.audio_enabled:
            # send audio at about the same rate as video
            # this can be changed
            self.last_audio = (np.zeros((int(AUDIORATE/self.fps), )),
                               np.zeros((int(AUDIORATE/self.fps), )))
            self.last_audio_time = None
            self.next_audio_send_time = None
            self.audio_frame_counter = 0
            self.q_audio = queue.PriorityQueue()
            self.t = threading.Timer(0.0, self._send_audio)
            self.t.daemon = True
            self.t.start()

    def _send_video_frame(self):
        start_time = time.time()
        try:
            frame = self.q_video.get_nowait()
            # frame[0] is frame count of the frame
            # frame[1] is the frame
            frame = frame[1]
        except IndexError:
            frame = self.last_frame
        except queue.Empty:
            frame = self.last_frame
        else:
            self.last_frame = frame

        try:
            super(TwitchBufferedOutputStream, self
                  ).send_video_frame(frame)
        except OSError:
            # stream has been closed.
            # This function is still called once when that happens.
            # Don't call this function again and everything should be
            # cleaned up just fine.
            return

        # send the next frame at the appropriate time
        if self.next_video_send_time is None:
            self.t = threading.Timer(1./self.fps, self._send_video_frame)
            self.next_video_send_time = start_time + 1./self.fps
        else:
            self.next_video_send_time += 1./self.fps
            next_event_time = self.next_video_send_time - start_time
            if next_event_time > 0:
                self.t = threading.Timer(next_event_time,
                                         self._send_video_frame)
            else:
                # we should already have sent something!
                #
                # not allowed for recursion problems :-(
                # (maximum recursion depth)
                # self.send_me_last_frame_again()
                #
                # other solution:
                self.t = threading.Thread(
                    target=self._send_video_frame)

        self.t.daemon = True
        self.t.start()

    def _send_audio(self):
        start_time = time.time()
        try:
            _, left_audio, right_audio = self.q_audio.get_nowait()
        except IndexError:
            left_audio, right_audio = self.last_audio
        except queue.Empty:
            left_audio, right_audio = self.last_audio
        else:
            self.last_audio = (left_audio, right_audio)

        try:
            super(TwitchBufferedOutputStream, self
                  ).send_audio(left_audio, right_audio)
        except OSError:
            # stream has been closed.
            # This function is still called once when that happens.
            # Don't call this function again and everything should be
            # cleaned up just fine.
            return

        # send the next frame at the appropriate time
        downstream_time = len(left_audio) / AUDIORATE

        if self.next_audio_send_time is None:
            self.t = threading.Timer(downstream_time,
                                     self._send_audio)
            self.next_audio_send_time = start_time + downstream_time
        else:
            self.next_audio_send_time += downstream_time
            next_event_time = self.next_audio_send_time - start_time
            if next_event_time > 0:
                self.t = threading.Timer(next_event_time,
                                         self._send_audio)
            else:
                # we should already have sent something!
                #
                # not allowed for recursion problems :-(
                # (maximum recursion depth)
                # self.send_me_last_frame_again()
                #
                # other solution:
                self.t = threading.Thread(
                    target=self._send_audio)

        self.t.daemon = True
        self.t.start()

    def send_video_frame(self, frame, frame_counter=None):
        """send frame of shape (height, width, 3)
        with values between 0 and 1

        :param frame: array containing the frame.
        :type frame: numpy array with shape (height, width, 3)
            containing values between 0.0 and 1.0
        :param frame_counter: frame position number within stream.
            Provide this when multi-threading to make sure frames don't
            switch position
        :type frame_counter: int
        """
        if frame_counter is None:
            frame_counter = self.frame_counter
            self.frame_counter += 1

        self.q_video.put((frame_counter, frame))

    def send_audio(self,
                   left_channel,
                   right_channel,
                   frame_counter=None):
        """Add the audio samples to the stream. The left and the right
        channel should have the same shape.

        :param left_channel: array containing the audio signal.
        :type left_channel: numpy array with shape (k, )
            containing values between -1.0 and 1.0. l can be any integer
        :param right_channel: array containing the audio signal.
        :type right_channel: numpy array with shape (k, )
            containing values between -1.0 and 1.0. l can be any integer
        :param frame_counter: frame position number within stream.
            Provide this when multi-threading to make sure frames don't
            switch position
        :type frame_counter: int
        """
        if frame_counter is None:
            frame_counter = self.audio_frame_counter
            self.audio_frame_counter += 1

        self.q_audio.put((frame_counter, left_channel, right_channel))

    def get_video_frame_buffer_state(self):
        """Find out how many video frames are left in the buffer.
        The buffer should never run dry, or audio and video will go out
        of sync. Likewise, the more filled the buffer, the higher the
        memory use and the delay between you putting your frame in the
        stream and the frame showing up on Twitch.

        :return integer estimate of the number of video frames left.
        """
        return self.q_video.qsize()

    def get_audio_buffer_state(self):
        """Find out how many audio fragments are left in the buffer.
        The buffer should never run dry, or audio and video will go out
        of sync. Likewise, the more filled the buffer, the higher the
        memory use and the delay between you putting your frame in the
        stream and the frame showing up on Twitch.

        :return integer estimate of the number of audio fragments left.
        """
        return self.q_audio.qsize()




#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
The setup file for installing the library
"""
import os
from setuptools import find_packages
from setuptools import setup
from twitchstream import __version__ as VERSION
version = VERSION

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
    CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()
except IOError:
    README = CHANGES = ''



install_requires = [
    'numpy',
    ]

tests_require = [
    'mock',
    'pytest',
    'pytest-cov',
    'pytest-pep8',
    ]

setup(
    name="twitch-stream.py",
    version=version,
    description="An interface to the Twitch website, to interact with "
                "their video and chat",
    long_description="\n\n".join([README, CHANGES]),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Topic :: Communications :: Chat :: Internet Relay Chat",
        "Topic :: Multimedia :: Video"
        ],
    keywords="twitch, stream, video, chat",
    author="Jonas Degrave",
    author_email="erstaateenknolraapinmijntuin+pythontwitch@gmail.com",
    url="https://github.com/317070/python-twitch-stream",
    license="MIT",
    packages=find_packages(),
    include_package_data=False,
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
        'testing': tests_require,
        },
    )
